library unify.impl.common.http;

import 'dart:async';

import 'package:intl/intl.dart';
import 'package:meta/meta.dart';

import '../../../unify.dart';

abstract class UnifyHttpClient implements UnifyClient {
  UnifyHttpClient(this.email, this.password, this.token, this.endpoints);

  @override
  final String email;

  @override
  final String password;

  @override
  final UnifyApiEndpoints endpoints;

  @override
  UnifyAccessToken token;

  bool isCurrentTokenValid(bool refreshable) {
    if (token == null) {
      return false;
    }

    if (refreshable) {
      var now = DateTime.now();
      return token.expiresAt.difference(now).abs().inSeconds >= 60;
    }

    var now = DateTime.now();
    return token.expiresAt.difference(now).abs().inSeconds >= 60;
  }

  @override
  bool get isAuthorized => isCurrentTokenValid(false);

  @override
  Future login() async {
    if (!isCurrentTokenValid(false)) {
      var result = await postHttp(
        "/token",
        body: {
          "grant_type": "password",
          // "client_id": endpoints.clientId,
          // "client_secret": endpoints.clientSecret,
          "username": email,
          "password": password
        },
      );

      token = UnifyJsonAccessToken(result);
      return;
    }

    // var result = await loginHttp(
    //   "/token",
    //   body: {
    //     "grant_type": "refresh_token",
    //     // "client_id": endpoints.clientId,
    //     // "client_secret": endpoints.clientSecret,
    //     "refresh_token": token.refreshToken
    //   },
    //   contentType: "application/x-www-form-urlencoded",
    // );

    // token = UnifyJsonAccessToken(result);
  }

  @override
  Future setToken({UnifyJsonAccessToken accessToken}) {
    token = accessToken;
  }

  @override
  Future logout() async {
    token = null;
    return;
  }

  @override
  Future<String> getAccountStatus({String username, String password}) async {
    var _data = {
      "Username": username,
      "Password": password,
    };
    var result = await getJsonMap("account/status", body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to get status '${username}': ${reason}");
      } else {
        throw Exception("Failed to get status '${username}'");
      }
    }
    return result["Message"]?.toString();
  }

  Future<List<CompanyCategory>> listCompanyCategories() async {
    var items = <CompanyCategory>[];

    List<dynamic> result;

    result = await getJsonList("contacts/info/company_categories");

    for (var item in result) {
      items.add(CompanyCategory(this, item));
    }

    return items;
  }

  @override
  Future<List<Contact>> listContacts(
      {int page = 1,
      int rows = 100,
      String query = "",
      bool all = false}) async {
    var items = <Contact>[];

    List<dynamic> result;

    if (all) {
      page = 1;
      try {
        final _total = await _getCount(
            "contacts/1/1?First_Name=$query&Last_Name=$query&Email_Address=$query&Loan_Number=$query&Search_All=true");
        rows = (_total > 0) ? _total : rows;
      } catch (e) {
        rows = 1;
      }
    }

    if (query != null && query.isNotEmpty) {
      result = await getJsonList(
          "contacts/$rows/$page?First_Name=$query&Last_Name=$query&Email_Address=$query&Loan_Number=$query&Search_All=true");
    } else {
      result = await getJsonList("contacts/$rows/$page");
    }

    for (var item in result) {
      items.add(Contact(this, item));
    }

    return items;
  }

  @override
  Future<AccountStats> getTotals() async {
    int _contacts = await _getCount('contacts/1/1');
    int _leads = await _getCount('leads/1/1');
    int _loans = await _getCount('loans/1/1');
    int _loansMyPipeline = await _getCount('loans/1/1/mypipeline');
    int _loansActiveLoans = await _getCount('loans/1/1/activeloans');
    int _loansProspects = await _getCount('loans/1/1/prospects');
    int _loansClosed = await _getCount('loans/1/1/closed');
    int _loansOther = await _getCount('loans/1/1/other');
    int _leadsCore = await _getCount('core/leads/1/1');
    int _dashboardPartners = await _getCount('dashboard/1/1/partners');
    int _dashboardPipeline = await _getCount('dashboard/1/1/pipeline');
    int _dashboardCustomer = await _getCount('dashboard/1/1/customer');
    int _dashboardHotProspect = await _getCount('dashboard/1/1/hotprospect');
    return AccountStats(
      contacts: _contacts,
      leads: _leads,
      loans: _loans,
      loansMyPipeline: _loansMyPipeline,
      loansActiveLoans: _loansActiveLoans,
      loansProspects: _loansProspects,
      loansClosed: _loansClosed,
      loansOther: _loansOther,
      coreLeads: _leadsCore,
      dashboardPartners: _dashboardPartners,
      dashboardCustomer: _dashboardCustomer,
      dashboardHotProspect: _dashboardHotProspect,
      dashboardPipeline: _dashboardPipeline,
    );
  }

  Future<int> _getCount(String path) async {
    try {
      final _result = await getJsonMap("$path", extract: 'Paging');
      int _items = _result['totalCount'] ?? 0;
      return _items;
    } catch (e) {
      print(e);
      return 0;
    }
  }

  @override
  Future<ContactDetails> getContactDetails(String id) async {
    var _data = await getJsonMap("contacts/details/$id");
    if (_data == null) throw Exception("Contact Not Found");
    // List<ContactNote> _notes = await listContactNotes(id) ?? [];
    // List<Timeline> _tracker = await listContactTimeline(id) ?? [];
    // List<Task> _tasks = await listTasksForContact(id) ?? [];
    // List<LoanHistory> _loans = await listLoanHistory(id) ?? [];
    // _data["Contact_Notes"] = _notes?.map((i) => i?.json)?.toList() ?? [];
    // _data["Contact_Timeline"] = _tracker?.map((i) => i?.json)?.toList() ?? [];
    // _data["Contact_Tasks"] = _tasks?.map((i) => i?.json)?.toList() ?? [];
    // _data["Contact_Loan_History"] = _loans?.map((i) => i?.json)?.toList() ?? [];
    return ContactDetails(this, _data, id);
  }

  @override
  Future<List<ContactNote>> listContactNotes(String id) async {
    var items = <ContactNote>[];

    List<dynamic> result;

    result = await getJsonList("contacts/details/$id/notes");

    for (var item in result) {
      items.add(ContactNote(this, item));
    }

    return items;
  }

  @override
  Future<List<Lead>> listLeads(
      {int page = 1,
      int rows = 100,
      String query = "",
      bool all = false}) async {
    var items = <Lead>[];

    List<dynamic> result;

    if (all) {
      page = 1;
      try {
        final _total = await _getCount(
            "leads/1/1?First_Name=$query&Last_Name=$query&Email_Address=$query&Loan_Number=$query&Search_All=true");
        rows = (_total > 0) ? _total : rows;
      } catch (e) {
        rows = 0;
      }
    }

    if (query != null && query.isNotEmpty) {
      result = await getJsonList(
          "leads/$rows/$page?First_Name=$query&Last_Name=$query&Email_Address=$query&Loan_Number=$query&Search_All=true");
    } else {
      result = await getJsonList("leads/$rows/$page");
    }

    for (var item in result) {
      items.add(Lead(this, item));
    }

    return items;
  }

  @override
  Future<LeadDetails> getLeadDetails(String id) async {
    var _data = await getJsonMap("leads/details/$id");
    if (_data == null) throw Exception("Lead Not Found");
    // List<LeadNote> _notes = await listLeadNotes(id) ?? [];
    // List<Timeline> _tracker = await listLeadTimeline(id) ?? [];
    // List<Task> _tasks = await listTasksForLead(id) ?? [];
    // _data["Lead_Notes"] = _notes?.map((i) => i?.json)?.toList() ?? [];
    // _data["Lead_Timeline"] = _tracker?.map((i) => i?.json)?.toList() ?? [];
    // _data["Lead_Tasks"] = _tasks?.map((i) => i?.json)?.toList() ?? [];
    return LeadDetails(this, _data, id);
  }

  @override
  Future<List<LeadNote>> listLeadNotes(String id) async {
    var items = <LeadNote>[];

    List<dynamic> result;

    result = await getJsonList("leads/details/$id/notes");

    for (var item in result) {
      items.add(LeadNote(this, item));
    }

    return items;
  }

  @override
  Future editContact(
    String id, {
    @required String firstName,
    @required String lastName,
    String middleName,
    String suffix,
    String secondaryFirstName,
    String secondaryMiddleName,
    String secondaryLastName,
    PostalAddress currentAddress,
    PhoneNumber cellPhone,
    PhoneNumber homePhone,
    PhoneNumber officePhone,
    PhoneNumber faxPhone,
    PhoneNumber secondaryCellPhone,
    String email,
    String secondaryEmail,
    String loanNumber,
    int contactPriority,
    CompanyCategory companyCategory,
    List<ContactGroup> groups,
  }) async {
    ContactDetails _info = await getContactDetails(id);

    var _data = {
      "First_Name": firstName ?? _info?.firstName,
      "Middle_Name": middleName ?? _info?.middleName,
      "Last_Name": lastName ?? _info?.lastName,
      "Suffix": suffix ?? _info?.suffix,
      "Loan_Number": loanNumber ?? _info?.loanNumber,
      "Contact_Priority": contactPriority ?? _info?.priority ?? 0,
      "Secondary_First_Name": secondaryFirstName ?? _info?.secondaryFirstName,
      "Secondary_Middle_Name":
          secondaryMiddleName ?? _info?.secondaryMiddleName,
      "Secondary_Last_Name": secondaryLastName ?? _info?.secondaryLastName,
      "Address": currentAddress?.json ?? _info?.currentAddress?.json,
      "Home_Phone": homePhone?.json ?? _info?.cellPhone?.json,
      "Office_Phone": officePhone?.json ?? _info?.officePhone?.json,
      "Cell_Phone": cellPhone?.json ?? _info?.homePhone?.json,
      "Secondary_Cell_Phone":
          secondaryCellPhone?.json ?? _info?.secondaryCellPhone?.json,
      "Fax_Phone": faxPhone?.json ?? _info?.faxPhone?.json,
      "Email_Address": email ?? _info?.email ?? "",
      "Secondary_Email_Address": secondaryEmail ?? _info?.secondaryEmail ?? "",
      "Company_Category": companyCategory?.json ?? _info?.companyCategory?.json,
      "Contact_Groups": groups?.map((g) => g.json)?.toList() ??
          _info?.groups?.map((g) => g.json)?.toList() ??
          [],
    };
    var result = await getJsonMap("contacts/details/$id",
        body: _data, extract: null, method: "PUT");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to edit contact '${firstName}': ${reason}");
      } else {
        throw Exception("Failed to edit contact '${firstName}'");
      }
    }
  }

  @override
  Future addContact({
    @required String firstName,
    @required String lastName,
    String middleName,
    String suffix,
    String secondaryFirstName,
    String secondaryMiddleName,
    String secondaryLastName,
    PostalAddress currentAddress,
    PhoneNumber cellPhone,
    PhoneNumber homePhone,
    PhoneNumber officePhone,
    PhoneNumber faxPhone,
    PhoneNumber secondaryCellPhone,
    String email,
    String secondaryEmail,
    String loanNumber,
    int contactPriority,
    CompanyCategory companyCategory,
    List<ContactGroup> groups,
  }) async {
    var _data = {
      "First_Name": firstName,
      "Middle_Name": middleName ?? "",
      "Last_Name": lastName,
      "Suffix": suffix ?? "",
      "Loan_Number": loanNumber ?? "",
      "Contact_Priority": contactPriority ?? 0,
      "Secondary_First_Name": secondaryFirstName ?? "",
      "Secondary_Middle_Name": secondaryMiddleName ?? "",
      "Secondary_Last_Name": secondaryLastName ?? "",
      "Address": currentAddress?.json,
      "Home_Phone": homePhone?.json,
      "Office_Phone": officePhone?.json,
      "Cell_Phone": cellPhone?.json,
      "Secondary_Cell_Phone": secondaryCellPhone?.json,
      "Fax_Phone": faxPhone?.json,
      "Email_Address": email ?? "",
      "Secondary_Email_Address": secondaryEmail ?? "",
      "Company_Category": companyCategory?.json,
      "Contact_Groups": groups?.map((g) => g.json)?.toList() ?? [],
    };
    var result = await getJsonMap("contacts/add", body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add contact '${firstName}': ${reason}");
      } else {
        throw Exception("Failed to add contact '${firstName}'");
      }
    }
  }

  @override
  Future editLead(
    String id, {
    @required String firstName,
    @required String lastName,
    String middleName,
    String suffix,
    String secondaryFirstName,
    String secondaryMiddleName,
    String secondaryLastName,
    PostalAddress currentAddress,
    PostalAddress propertyAddress,
    PhoneNumber cellPhone,
    PhoneNumber homePhone,
    PhoneNumber officePhone,
    PhoneNumber faxPhone,
    PhoneNumber secondaryCellPhone,
    String email,
    List<LeadGroup> groups,
    String bestCallTime,
    String creditRange,
    String secondaryCreditRange,
    double annualGrossIncome,
    bool recentBankruptcy,
    double propertyValue,
  }) async {
    LeadDetails _info = await getLeadDetails(id);

    var _data = {
      "First_Name": firstName ?? _info?.firstName,
      "Middle_Name": middleName ?? _info?.middleName,
      "Last_Name": lastName ?? _info?.lastName,
      "Suffix": suffix ?? _info?.suffix,
      "Secondary_First_Name": secondaryFirstName ?? _info?.secondaryFirstName,
      "Secondary_Middle_Name":
          secondaryMiddleName ?? _info?.secondaryMiddleName,
      "Secondary_Last_Name": secondaryLastName ?? _info?.secondaryLastName,
      "Current_Address": currentAddress?.json ?? _info?.currentAddress?.json,
      "Property_Address": propertyAddress?.json ?? _info?.propertyAddress?.json,
      "Home_Phone": homePhone?.json ?? _info?.cellPhone?.json,
      "Office_Phone": officePhone?.json ?? _info?.officePhone?.json,
      "Cell_Phone": cellPhone?.json ?? _info?.homePhone?.json,
      "Secondary_Cell_Phone":
          secondaryCellPhone?.json ?? _info?.secondaryCellPhone?.json,
      "Fax_Phone": faxPhone?.json ?? _info?.faxPhone?.json,
      "Email_Address": email ?? _info?.email ?? "",
      "Lead_Groups": groups?.map((g) => g.json)?.toList() ??
          _info?.groups?.map((g) => g.json)?.toList() ??
          [],
      "Best_Call_Time": bestCallTime ?? _info?.bestCallTime ?? "",
      "Credit_Range": creditRange ?? _info?.creditRange ?? "",
      "Secondary_Credit_Range":
          secondaryCreditRange ?? _info?.secondaryCreditRange ?? "",
      "Annual_Gross_Income": annualGrossIncome ?? _info?.annualGrossIncome ?? 0,
      "Recent_Bankruptcy": recentBankruptcy ?? _info?.recentBankruptcy ?? false,
      "Property_Value": propertyValue ?? _info?.propertyValue ?? 0,
    };
    var result = await getJsonMap("leads/details/$id",
        body: _data, extract: null, method: "PUT");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to edit lead '${firstName}': ${reason}");
      } else {
        throw Exception("Failed to edit lead '${firstName}'");
      }
    }
  }

  @override
  Future addLead({
    @required String firstName,
    @required String lastName,
    String middleName,
    String suffix,
    String secondaryFirstName,
    String secondaryMiddleName,
    String secondaryLastName,
    PostalAddress currentAddress,
    PostalAddress propertyAddress,
    PhoneNumber cellPhone,
    PhoneNumber homePhone,
    PhoneNumber officePhone,
    PhoneNumber faxPhone,
    PhoneNumber secondaryCellPhone,
    String email,
    List<LeadGroup> groups,
    String bestCallTime,
    String creditRange,
    String secondaryCreditRange,
    double annualGrossIncome,
    bool recentBankruptcy,
    double propertyValue,
  }) async {
    var _data = {
      "First_Name": firstName,
      "Middle_Name": middleName ?? "",
      "Last_Name": lastName,
      "Suffix": suffix ?? "",
      "Secondary_First_Name": secondaryFirstName ?? "",
      "Secondary_Middle_Name": secondaryMiddleName ?? "",
      "Secondary_Last_Name": secondaryLastName ?? "",
      "Current_Address": currentAddress?.json,
      "Property_Address": propertyAddress?.json,
      "Home_Phone": homePhone?.json,
      "Office_Phone": officePhone?.json,
      "Cell_Phone": cellPhone?.json,
      "Secondary_Cell_Phone": secondaryCellPhone?.json,
      "Fax_Phone": faxPhone?.json,
      "Email_Address": email ?? "",
      "Lead_Groups": groups?.map((g) => g.json)?.toList() ?? [],
      "Best_Call_Time": bestCallTime ?? "",
      "Credit_Range": creditRange ?? "",
      "Secondary_Credit_Range": secondaryCreditRange ?? "",
      "Annual_Gross_Income": annualGrossIncome ?? 0,
      "Recent_Bankruptcy": recentBankruptcy ?? false,
      "Property_Value": propertyValue ?? 0,
    };
    var result = await getJsonMap("leads/add", body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add lead '${firstName}': ${reason}");
      } else {
        throw Exception("Failed to add lead '${firstName}'");
      }
    }
  }

  @override
  Future addContactNote(String id, String text) async {
    var _data = {"text": text ?? ""};
    var result = await getJsonMap("contacts/details/$id/notes",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add note '${text}': ${reason}");
      } else {
        throw Exception("Failed to add note '${text}'");
      }
    }
  }

  @override
  Future addLeadNote(String id, String text) async {
    var _data = {"text": text ?? ""};
    var result =
        await getJsonMap("leads/details/$id/notes", body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add note '${text}': ${reason}");
      } else {
        throw Exception("Failed to add note '${text}'");
      }
    }
  }

  @override
  Future addContactLogResponse(String id,
      {@required String action, String comment}) async {
    var _data = {
      "action": action,
      "comment": comment ?? "",
    };
    var result = await getJsonMap("contacts/details/$id/log_response",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add log response '${action}': ${reason}");
      } else {
        throw Exception("Failed to add log response '${action}'");
      }
    }
  }

  @override
  Future addLeadLogResponse(String id,
      {@required String action, String comment}) async {
    var _data = {
      "action": action,
      "comment": comment ?? "",
    };
    var result = await getJsonMap("leads/details/$id/log_response",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add log response '${action}': ${reason}");
      } else {
        throw Exception("Failed to add log response '${action}'");
      }
    }
  }

  @override
  Future addContactFollowUp(String id,
      {String title,
      DateTime date,
      int type = 0,
      String description = "",
      bool dismissOnChange = true,
      String alertType,
      int alertOffset = 0,
      String userNumber,
      String prospectNumber}) async {
    var _data = {
      "task_type": type ?? 0,
      "title": title,
      "description": description ?? "",
      "date_time": (date ?? DateTime.now()).toString(),
      "dismiss_on_change": dismissOnChange ?? true,
      "alert_type": alertType ?? "",
      "alert_offset": alertOffset ?? 0,
      "prospect_number": userNumber,
      "user_number": prospectNumber
    };
    var result = await getJsonMap("contacts/details/$id/follow_up",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add follow up '${title}': ${reason}");
      } else {
        throw Exception("Failed to add follow up '${title}'");
      }
    }
  }

  @override
  Future addLeadFollowUp(String id,
      {String title,
      DateTime date,
      int type = 0,
      String description = "",
      bool dismissOnChange = true,
      String alertType,
      int alertOffset = 0,
      String userNumber,
      String prospectNumber}) async {
    var _data = {
      "task_type": type ?? 0,
      "title": title,
      "description": description ?? "",
      "date_time": (date ?? DateTime.now()).toString(),
      "dismiss_on_change": dismissOnChange ?? true,
      "alert_type": alertType ?? "",
      "alert_offset": alertOffset ?? 0,
      "prospect_number": userNumber,
      "user_number": prospectNumber
    };
    var result = await getJsonMap("leads/details/$id/follow_up",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add follow up '${title}': ${reason}");
      } else {
        throw Exception("Failed to add follow up '${title}'");
      }
    }
  }

  @override
  Future<List<Timeline>> listLeadTimeline(String id) async {
    var items = <Timeline>[];

    List<dynamic> result;

    result = await getJsonList("leads/details/$id/tracker");

    for (var item in result) {
      items.add(Timeline(this, item));
    }

    return items;
  }

  @override
  Future<List<Timeline>> listContactTimeline(String id) async {
    var items = <Timeline>[];

    List<dynamic> result;

    result = await getJsonList("contacts/details/$id/tracker");

    for (var item in result) {
      items.add(Timeline(this, item));
    }

    return items;
  }

  @override
  Future<List<InquiryAlert>> listInquiryAlerts(
      {int page = 1, int rows = 100, bool all = false}) async {
    var items = <InquiryAlert>[];

    List<dynamic> result;

    if (all) {
      result = await getJsonList("ServiceAlerts/$rows/$page/inquiry?all=true");
    } else {
      result = await getJsonList("ServiceAlerts/$rows/$page/inquiry?all=false");
    }

    for (var item in result) {
      items.add(InquiryAlert(this, item));
    }

    return items;
  }

  @override
  Future<List<DashboardAlert>> listOpportunitiyAlerts(
      {int page = 1, int rows = 100, bool all = false}) async {
    var items = <DashboardAlert>[];

    List<dynamic> result;

    if (all) {
      page = 1;
      try {
        final _total = await _getCount("dashboard/1/1/opportunity?all=true");
        rows = (_total > 0) ? _total : rows;
      } catch (e) {
        rows = 1;
      }
    }

    result = await getJsonList("dashboard/$rows/$page/opportunity");

    for (var item in result) {
      items.add(DashboardAlert(this, item));
    }

    return items;
  }

  @override
  Future<List<DashboardAlert>> listCustomerAlerts(
      {int page = 1, int rows = 100, bool all = false}) async {
    var items = <DashboardAlert>[];

    List<dynamic> result;

    if (all) {
      page = 1;
      try {
        final _total = await _getCount("dashboard/1/1/customer?all=true");
        rows = (_total > 0) ? _total : rows;
      } catch (e) {
        rows = 1;
      }
    }

    result = await getJsonList("dashboard/$rows/$page/customer");

    for (var item in result) {
      items.add(DashboardAlert(this, item));
    }

    return items;
  }

  @override
  Future<List<DashboardAlert>> listHotProspectAlerts(
      {int page = 1, int rows = 100, bool all = false}) async {
    var items = <DashboardAlert>[];

    List<dynamic> result;

    if (all) {
      page = 1;
      try {
        final _total = await _getCount("dashboard/1/1/hotprospect?all=true");
        rows = (_total > 0) ? _total : rows;
      } catch (e) {
        rows = 1;
      }
    }

    result = await getJsonList("dashboard/$rows/$page/hotprospect");

    for (var item in result) {
      items.add(DashboardAlert(this, item));
    }

    return items;
  }

  @override
  Future<List<DashboardAlert>> listPipelineAlerts(
      {int page = 1, int rows = 100, bool all = false}) async {
    var items = <DashboardAlert>[];

    List<dynamic> result;

    if (all) {
      page = 1;
      try {
        final _total = await _getCount("dashboard/1/1/pipeline?all=true");
        rows = (_total > 0) ? _total : rows;
      } catch (e) {
        rows = 1;
      }
    }

    result = await getJsonList("dashboard/$rows/$page/pipeline");

    for (var item in result) {
      items.add(DashboardAlert(this, item));
    }

    return items;
  }

  @override
  Future<List<DashboardPartners>> listDashboardPartners(
      {int page = 1, int rows = 100, bool all = false}) async {
    var items = <DashboardPartners>[];

    List<dynamic> result;

    if (all) {
      page = 1;
      try {
        final _total = await _getCount("dashboard/1/1/partners?all=true");
        rows = (_total > 0) ? _total : rows;
      } catch (e) {
        rows = 1;
      }
    }

    result = await getJsonList("dashboard/$rows/$page/partners");

    for (var item in result) {
      items.add(DashboardPartners(this, item));
    }

    return items;
  }

  @override
  Future<List<ContactGroup>> listContactGroups({bool totals = true}) async {
    var items = <ContactGroup>[];

    List<dynamic> result;

    result = await getJsonList("contacts/contact_groups?totals=$totals");

    for (var item in result) {
      items.add(ContactGroup(this, item));
    }

    return items;
  }

  @override
  Future<List<ContactDetails>> addContactsToGroup(
      {List<String> contactIDs, String groupID}) async {
    var items = <ContactDetails>[];
    if (groupID != null || groupID.isNotEmpty) {
      if (contactIDs.isNotEmpty) {
        var result = await getJsonList(
            "contacts/contact_groups/add/batch?id=$groupID",
            strBody: contactIDs,
            method: "POST");
        for (int i = 0; i < result.length; i++) {
          var item = result[i];
          items.add(ContactDetails(this, item, contactIDs[i]));
        }
        return items;
      } else {
        throw Exception("Failed: Please provide at least 1 contactID");
      }
    } else {
      throw Exception("Failed: Please provide a group ID");
    }
  }

  @override
  Future<List<LeadDetails>> addLeadsToGroup(
      {List<String> leadIDs, String groupID}) async {
    var items = <LeadDetails>[];
    if (groupID != null || groupID.isNotEmpty) {
      if (leadIDs.isNotEmpty) {
        var result = await getJsonList(
            "/api/leads/lead_groups/add/batch?id=$groupID",
            strBody: leadIDs,
            method: "POST");
        for (int i = 0; i < result.length; i++) {
          var item = result[i];
          items.add(LeadDetails(this, item, leadIDs[i]));
        }
        return items;
      } else {
        throw Exception("Failed: Please provide at least 1 contactID");
      }
    } else {
      throw Exception("Failed: Please provide a group ID");
    }
  }

  @override
  Future<List<LeadGroup>> listLeadGroups({bool totals = true}) async {
    var items = <LeadGroup>[];

    List<dynamic> result;

    result = await getJsonList("leads/lead_groups?totals=$totals");

    for (var item in result) {
      items.add(LeadGroup(this, item));
    }

    return items;
  }

  @override
  Future<List<Contact>> listContactsForContactGroup(String id,
      {int page = 1, int rows = 100, String query = ""}) async {
    var items = <Contact>[];

    List<dynamic> result;

    if (query != null && query.isNotEmpty) {
      result = await getJsonList(
          "contacts/contact_groups/$id/list?rows=$rows&page=$page&First_Name=$query&Last_Name=$query&Email_Address=$query&Loan_Number=$query&Search_All=true");
    } else {
      result = await getJsonList(
          "contacts/contact_groups/$id/list?rows=$rows&page=$page");
    }

    for (var item in result) {
      items.add(Contact(this, item));
    }

    return items;
  }

  @override
  Future<List<Lead>> listLeadsForLeadGroup(String id,
      {int page = 1, int rows = 100, String query = ""}) async {
    var items = <Lead>[];

    List<dynamic> result;

    if (query != null && query.isNotEmpty) {
      result = await getJsonList(
          "leads/lead_groups/$id/list?rows=$rows&page=$page&First_Name=$query&Last_Name=$query&Email_Address=$query&Loan_Number=$query&Search_All=true");
    } else {
      result =
          await getJsonList("leads/lead_groups/$id/list?rows=$rows&page=$page");
    }

    for (var item in result) {
      items.add(Lead(this, item));
    }

    return items;
  }

  @override
  Future<List<String>> listEmailsForContactGroup(String id) async {
    var items = <String>[];

    List<dynamic> result;

    result = await getJsonList("contacts/contact_groups/$id/emails");

    for (var item in result) {
      items.add(item as String);
    }

    return items;
  }

  @override
  Future<List<String>> listEmailsForLeadGroup(String id) async {
    var items = <String>[];

    List<dynamic> result;

    result = await getJsonList("leads/lead_groups/$id/emails");

    for (var item in result) {
      items.add(item as String);
    }

    return items;
  }

  @override
  Future<List<Loan>> listLoans(
      {int page = 1,
      int rows = 100,
      String query = "",
      LoansFolder folder = LoansFolder.all,
      bool all = false}) async {
    var items = <Loan>[];

    List<dynamic> result;

    if (all) {
      page = 1;
      try {
        final _total = await _getCount(
            "loans/1/1?First_Name=$query&Last_Name=$query&Email_Address=$query&Loan_Number=$query&Search_All=true");
        rows = (_total > 0) ? _total : rows;
      } catch (e) {
        rows = 1;
      }
    }

    String _folder = '';
    if (folder != LoansFolder.all) {
      _folder = '/' + folder.toString().replaceAll('LoansFolder.', '');
    }
    if (query != null && query.isNotEmpty) {
      result = await getJsonList(
          "loans/$rows/$page$_folder?First_Name=$query&Last_Name=$query&Email_Address=$query&Loan_Number=$query&Search_All=true");
    } else {
      result = await getJsonList("loans/$rows/$page$_folder");
    }

    for (var item in result) {
      items.add(Loan(this, item));
    }

    return items;
  }

  @override
  Future<LoanDetails> getLoanDetails(String id) async {
    var _data = await getJsonMap("loans/info/$id");
    if (_data == null) throw Exception("Loan Not Found");
    LoanProgress _progress = await getLoanStatus(id);
    List<LoanContact> _contacts = await listLoanContacts(id) ?? [];
    _data["loan_contacts"] = _contacts?.map((i) => i?.json)?.toList() ?? [];
    _data["loan_progress"] = _progress?.json;
    return LoanDetails(this, _data, id);
  }

  String get _calendarDateFormat => "MM-dd-yyyy";

  @override
  Future<List<Task>> listTasksContacts({DateTime date, int month}) async {
    String _date =
        DateFormat(_calendarDateFormat).format(date ?? DateTime.now());
    var items = <Task>[];

    var url = 'calendar/contact/$_date';

    if (month != null && month > 0 && month <= 12) {
      url += '?month=$month';
    }

    List<dynamic> result;

    result = await getJsonList(url);

    for (var item in result) {
      items.add(Task(this, item));
    }

    return items;
  }

  @override
  Future<List<Task>> listTasksLeads({DateTime date, int month}) async {
    String _date =
        DateFormat(_calendarDateFormat).format(date ?? DateTime.now());
    var items = <Task>[];

    var url = 'calendar/lead/$_date';

    if (month != null && month > 0 && month <= 12) {
      url += '?month=$month';
    }

    List<dynamic> result;

    result = await getJsonList(url);

    for (var item in result) {
      items.add(Task(this, item));
    }

    return items;
  }

  @override
  Future<List<Task>> listTasksAll({DateTime date, int month}) async {
    String _date =
        DateFormat(_calendarDateFormat).format(date ?? DateTime.now());
    var items = <Task>[];

    var url = 'calendar/$_date';

    if (month != null && month > 0 && month <= 12) {
      url += '?month=$month';
    }

    List<dynamic> result;

    result = await getJsonList(url);

    for (var item in result) {
      items.add(Task(this, item));
    }

    return items;
  }

  @override
  Future<List<Task>> listTasksCoreLeads({DateTime date, int month}) async {
    String _date =
        DateFormat(_calendarDateFormat).format(date ?? DateTime.now());
    var items = <Task>[];

    var url = 'calendar/corelead/$_date';

    if (month != null && month > 0 && month <= 12) {
      url += '?month=$month';
    }

    List<dynamic> result;

    result = await getJsonList(url);

    for (var item in result) {
      items.add(Task(this, item));
    }

    return items;
  }

  @override
  Future<List<Task>> listTasksForCoreLead(String id,
      {DateTime date, int month}) async {
    String _date =
        DateFormat(_calendarDateFormat).format(date ?? DateTime.now());
    var items = <Task>[];

    var url = 'calendar/corelead/$id/$_date';

    if (month != null && month > 0 && month <= 12) {
      url += '?month=$month';
    }

    List<dynamic> result;

    result = await getJsonList(url);

    for (var item in result) {
      items.add(Task(this, item));
    }

    return items;
  }

  @override
  Future<List<Task>> listTasksForContact(String id,
      {DateTime date, int month}) async {
    String _date =
        DateFormat(_calendarDateFormat).format(date ?? DateTime.now());
    var items = <Task>[];

    var url = 'calendar/contact/$id/$_date';

    if (month != null && month > 0 && month <= 12) {
      url += '?month=$month';
    }

    List<dynamic> result;

    result = await getJsonList(url);

    for (var item in result) {
      items.add(Task(this, item));
    }

    return items;
  }

  @override
  Future<List<Task>> listTasksForLead(String id,
      {DateTime date, int month}) async {
    String _date =
        DateFormat(_calendarDateFormat).format(date ?? DateTime.now());
    var items = <Task>[];

    var url = 'calendar/lead/$id/$_date';

    if (month != null && month > 0 && month <= 12) {
      url += '?month=$month';
    }

    List<dynamic> result;

    result = await getJsonList(url);

    for (var item in result) {
      items.add(Task(this, item));
    }

    return items;
  }

  @override
  Future<List<PreQual>> listPreQualTracker(
      {int page = 1,
      int rows = 100,
      String loanType = '',
      String query = '',
      String loanView = 'True'}) async {
    var items = <PreQual>[];
    List<dynamic> result;
    String typeString = '';

    if (loanType != null && loanType.isNotEmpty) {
      typeString = 'loan_Type=$loanType';
    }

    if (query != null && query.isNotEmpty) {
      result = await getJsonList(
          "GreatnessTracker/PreQual/$rows/$page?$typeString&file_Contact_Type=$query&contact_Email=$query&loan_View=$loanView");
    } else {
      result = await getJsonList(
          "GreatnessTracker/PreQual/$rows/$page?$typeString&loan_View=$loanView");
    }

    for (var item in result) {
      items.add(PreQual(this, item));
    }

    return items;
  }

  @override
  Future updateLoanView({String loanID = '', String loanView = ''}) async {
    var result = await getJsonMap(
        "GreatnessTracker/PreQual/UpdateLoanView?loan_ID=$loanID&loan_View=$loanView",
        extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to update: ${reason}");
      } else {
        throw Exception("Failed to update");
      }
    }
  }

  @override
  Future<List<StatusUpdate>> listStatusUpdateTracker({
    int page = 1,
    int rows = 100,
    String query = '',
    String weekDate = '',
  }) async {
    var items = <StatusUpdate>[];

    List<dynamic> result;

    if (query != null && query.isNotEmpty) {
      if (weekDate != null && weekDate.isNotEmpty) {
        result = await getJsonList(
            "GreatnessTracker/StatusUpdate/$rows/$page?file_Contact_Type=$query&contact_Email=$query&select_Week_Date=$weekDate");
      } else {
        result = await getJsonList(
            "GreatnessTracker/StatusUpdate/$rows/$page?file_Contact_Type=$query&contact_Email=$query");
      }
    } else {
      result = await getJsonList("GreatnessTracker/StatusUpdate/$rows/$page");
    }

    for (var item in result) {
      items.add(StatusUpdate(this, item));
    }

    return items;
  }

  @override
  Future addUpdateStatusTracker(
      {String loanID, String updateType = '', contactType = ''}) async {
    var result = await getJsonMap(
        "GreatnessTracker/StatusUpdate/AddUpdate?loan_ID=$loanID&update_Type=$updateType&contact_Type=$contactType",
        extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to update: ${reason}");
      } else {
        throw Exception("Failed to update");
      }
    }
  }

  @override
  Future addStatusUpdateResponse(
      {String contactID,
      String loanID,
      String contactType,
      String noteText,
      String noteType}) async {
    var _data = {
      "Contact_ID": contactID,
      "Loan_ID": loanID,
      "Contact_Type": contactType,
      "Note_Text": noteText,
      "Note_Type": noteType
    };
    var result = await getJsonMap(
        "GreatnessTracker/StatusUpdate/AddLogResponse",
        body: _data,
        extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add log response: ${reason}");
      } else {
        throw Exception("Failed to add log response");
      }
    }
  }

  @override
  Future<List<CoreLead>> listCoreLeads(
      {int page = 1, int rows = 100, String query = ""}) async {
    var items = <CoreLead>[];

    List<dynamic> result;

    if (query != null && query.isNotEmpty) {
      result = await getJsonList(
          "Core/leads/$rows/$page?First_Name=$query&Last_Name=$query&Email_Address=$query&Loan_Number=$query&Search_All=true");
    } else {
      result = await getJsonList("core/leads/$rows/$page");
    }

    for (var item in result) {
      items.add(CoreLead(this, item));
    }

    return items;
  }

  @override
  Future<List<AccountNames>> listActiveAccounts() async {
    var accounts = <AccountNames>[];

    var _data = await getJsonList("Core/AssignTo");
    if (_data == null) throw Exception("No Accounts Found");

    for (var item in _data) {
      accounts.add(AccountNames(this, item));
    }

    return accounts;
  }

  @override
  Future<CoreLeadDetails> getCoreLeadDetails(String id) async {
    var _data = await getJsonMap("Core/leads/details/$id");
    if (_data == null) throw Exception("Lead Not Found");
    List<Task> _tasks = await listTasksForLead(id) ?? [];
    _data["Lead_Timeline_Tasks"] = _tasks?.map((i) => i?.json)?.toList() ?? [];
    return CoreLeadDetails(this, _data, id);
  }

  @override
  Future addCoreLeadLogResponse(String id,
      {@required String action, String comment}) async {
    var _data = {
      "action": action,
      "comment": comment ?? "",
    };
    var result = await getJsonMap("Core/leads/$id/log_response",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add log response '${action}': ${reason}");
      } else {
        throw Exception("Failed to add log response '${action}'");
      }
    }
  }

  @override
  Future addCoreLead({
    bool attach = false,
    bool update = false,
    String attachContact,
    String updateID,
    String lead_Tracker_Lead_ID,
    String contact_ID,
    @required String borrower_First_Name,
    @required String borrower_Last_Name,
    PhoneNumber borrowerPhone1,
    PhoneNumber coBorrowerPhone1,
    PhoneNumber borrowerPhone2,
    PhoneNumber coBorrowerPhone2,
    PhoneNumber leadPhoneCell,
    PhoneNumber leadPhoneOffice,
    PhoneNumber agentPhoneCell,
    PhoneNumber agentPhoneOffice,
    String borrower_Birthdate,
    String borrower_Phone1_Type,
    String borrower_Phone2_Type,
    String borrower_Email_Address,
    String borrower_Contact_Preference,
    String borrower_Credit_Report_Auth,
    num borrower_Credit_Score,
    String coBorrower_First_Name,
    String coBorrower_Last_Name,
    String coBorrower_Birthdate,
    String coBorrower_Phone1_Type,
    String coBorrower_Phone2_Type,
    String coBorrower_Email_Address,
    String coBorrower_Contact_Preference,
    String coBorrower_Credit_Report_Auth,
    num coBorrower_Credit_Score,
    String hot_Lead,
    String next_Step,
    String next_Action_Date,
    String disposition,
    String assign_Lead_To,
    String initial_Appointment_Date,
    String appointment_Type,
    String mailing_Address1,
    String mailing_Address2,
    String mailing_City,
    String mailing_State,
    String mailing_Zip,
    String closing_Date,
    num interest_Rate,
    num loan_Amount,
    num appraised_Value,
    num purchase_Price,
    num monthly_Payments,
    num target_Refi_Rate,
    String loan_Type,
    String property_Type,
    String occupancy,
    String loan_Number,
    String loan_Purpose,
    String property_City,
    String property_State,
    String time_Frame,
    num property_Value,
    num request_Loan_Amount,
    num request_Purchase_Price,
    String down_Payment,
    String is_CoBorrower,
    String is_Military_Service,
    String details_Complete,
    String documentation_Complete,
    String encompass_File_Created,
    String builder_Loan,
    String lead_Source,
    String lead_Source_Type,
    String lead_Source_Name,
    String lead_Source_Email_Address,
    String lead_Source_Contact_ID,
    String sub_Division_Name,
    String new_Home_Sales_Agent,
    String sales_Agent_Email_Address,
    String date_PreApproved,
    String real_Deal_Date,
    String lead_Date,
    String lead_View,
    num total_count,
  }) async {
    Map<String, Object> _data;
    final defaultDate = "1/1/1800 12:00:00 AM";
    if (update) {
      final details = await getCoreLeadDetails(updateID);
      _data = {
        "Contact_ID": contact_ID ?? details.contactId,
        "Lead_Tracker_Lead_ID": lead_Tracker_Lead_ID ?? details.id,
        "Borrower_First_Name": borrower_First_Name ?? details.firstName,
        "Borrower_Last_Name": borrower_Last_Name ?? details.lastName,
        "Borrower_Birthdate": borrower_Birthdate ?? details.borrower_Birthdate,
        "Borrower_Phone1_Type": borrower_Phone1_Type ?? details.phoneType1,
        "Borrower_Phone1_Area_Code":
            borrowerPhone1?.areaCode ?? details.primaryPhone?.areaCode,
        "Borrower_Phone1_Prefix":
            borrowerPhone1?.prefix ?? details.primaryPhone?.prefix,
        "Borrower_Phone1_Number":
            borrowerPhone1?.number ?? details.primaryPhone?.number,
        "Borrower_Phone2_Type": borrower_Phone2_Type ?? details.phoneType2,
        "Borrower_Phone2_Area_Code":
            borrowerPhone2?.areaCode ?? details.secondaryPhone?.areaCode,
        "Borrower_Phone2_Prefix":
            borrowerPhone2?.prefix ?? details.secondaryPhone?.prefix,
        "Borrower_Phone2_Number":
            borrowerPhone2?.number ?? details.secondaryPhone?.number,
        "Borrower_Email_Address": borrower_Email_Address ?? details.email,
        "Borrower_Contact_Preference":
            borrower_Contact_Preference ?? details.borrower_Contact_Preference,
        "Borrower_Credit_Report_Auth":
            borrower_Credit_Report_Auth ?? details.borrower_Credit_Report_Auth,
        "Borrower_Credit_Score":
            borrower_Credit_Score ?? details.borrower_Credit_Score,
        "CoBorrower_First_Name": coBorrower_First_Name ?? details.cofirstName,
        "CoBorrower_Last_Name": coBorrower_Last_Name ?? details.colastName,
        "CoBorrower_Birthdate":
            coBorrower_Birthdate ?? details.coBorrower_Birthdate,
        "CoBorrower_Phone1_Type":
            coBorrower_Phone1_Type ?? details.coPhoneType1,
        "CoBorrower_Phone1_Area_Code":
            coBorrowerPhone1?.areaCode ?? details.coprimaryPhone?.areaCode,
        "CoBorrower_Phone1_Prefix":
            coBorrowerPhone1?.prefix ?? details.coprimaryPhone?.prefix,
        "CoBorrower_Phone1_Number":
            coBorrowerPhone1?.number ?? details.coprimaryPhone?.number,
        "CoBorrower_Phone2_Type":
            coBorrower_Phone2_Type ?? details.coPhoneType2,
        "CoBorrower_Phone2_Area_Code":
            coBorrowerPhone2?.areaCode ?? details.cosecondaryPhone?.areaCode,
        "CoBorrower_Phone2_Prefix":
            coBorrowerPhone2?.prefix ?? details.cosecondaryPhone?.prefix,
        "CoBorrower_Phone2_Number":
            coBorrowerPhone2?.number ?? details.cosecondaryPhone?.number,
        "CoBorrower_Email_Address": coBorrower_Email_Address ?? details.coemail,
        "CoBorrower_Contact_Preference": coBorrower_Contact_Preference ??
            details.coBorrower_Contact_Preference,
        "CoBorrower_Credit_Report_Auth": coBorrower_Credit_Report_Auth ??
            details.coBorrower_Credit_Report_Auth,
        "CoBorrower_Credit_Score":
            coBorrower_Credit_Score ?? details.coBorrower_Credit_Score,
        "Hot_Lead": hot_Lead ?? details.hot_Lead,
        "Next_Step": next_Step ?? details.next_Step,
        "Next_Action_Date": next_Action_Date ?? details.next_Action_Date,
        "Disposition": disposition ?? details.disposition,
        "Assign_Lead_To": assign_Lead_To ?? details.assign_Lead_To,
        "Initial_Appointment_Date":
            initial_Appointment_Date ?? details.initial_Appointment_Date,
        "Appointment_Type": appointment_Type ?? details.appointment_Type,
        "Mailing_Address1": mailing_Address1 ?? details.mailingAddress.street,
        "Mailing_Address2":
            mailing_Address2 ?? details.mailingAddress.apartment,
        "Mailing_City": mailing_City ?? details.mailingAddress.city,
        "Mailing_State": mailing_State ?? details.mailingAddress.state,
        "Mailing_Zip": mailing_Zip ?? details.mailingAddress.zip,
        "Closing_Date": closing_Date ?? details.closing_Date,
        "Interest_Rate": interest_Rate ?? details.interest_Rate,
        "Loan_Amount": loan_Amount ?? details.loan_Amount,
        "Appraised_Value": appraised_Value ?? details.appraised_Value,
        "Purchase_Price": purchase_Price ?? details.purchase_Price,
        "Monthly_Payments": monthly_Payments ?? details.monthly_Payments,
        "Target_Refi_Rate": target_Refi_Rate ?? details.target_Refi_Rate,
        "Loan_Type": loan_Type ?? details.loan_Type,
        "Property_Type": property_Type ?? details.property_Type,
        "Occupancy": occupancy ?? details.occupancy,
        "Loan_Number": loan_Number ?? details.loanNumber,
        "Loan_Purpose": loan_Purpose ?? details.loanPurpose,
        "Property_City": property_City ?? details.propertyCity,
        "Property_State": property_State ?? details.propertyState,
        "Time_Frame": time_Frame ?? details.time_Frame,
        "Property_Value": property_Value ?? details.property_Value,
        "Request_Loan_Amount":
            request_Loan_Amount ?? details.request_Loan_Amount,
        "Request_Purchase_Price":
            request_Purchase_Price ?? details.request_Purchase_Price,
        "Down_Payment": down_Payment ?? details.down_Payment,
        "Is_CoBorrower": is_CoBorrower ?? details.is_CoBorrower,
        "Details_Complete": details_Complete ?? details.details_Complete,
        "Documentation_Complete":
            documentation_Complete ?? details.documentation_Complete,
        "Encompass_File_Created":
            encompass_File_Created ?? details.encompass_File_Created,
        "Builder_Loan": builder_Loan ?? details.builder_Loan,
        "Lead_Source": lead_Source ?? details.leadSource,
        "Lead_Source_Type": lead_Source_Type ?? details.leadSourceType,
        "Lead_Source_Name": lead_Source_Name ?? details.leadSourceName,
        "Lead_Source_Office_Phone_Area_Code": leadPhoneOffice?.areaCode ??
            details.leadSourceOfficePhone?.areaCode,
        "Lead_Source_Office_Phone_Prefix":
            leadPhoneOffice?.prefix ?? details.leadSourceOfficePhone?.prefix,
        "Lead_Source_Office_Phone_Number":
            leadPhoneOffice?.number ?? details.leadSourceOfficePhone?.number,
        "Lead_Source_Cell_Phone_Area_Code":
            leadPhoneCell?.areaCode ?? details.leadSourceCellPhone?.areaCode,
        "Lead_Source_Cell_Phone_Prefix":
            leadPhoneCell?.prefix ?? details.leadSourceCellPhone?.prefix,
        "Lead_Source_Cell_Phone_Number":
            leadPhoneCell?.number ?? details.leadSourceCellPhone?.number,
        "Lead_Source_Email_Address":
            lead_Source_Email_Address ?? details.leadSourceEmail,
        "Lead_Source_Contact_ID":
            lead_Source_Contact_ID ?? details.lead_Source_Contact_ID,
        "Sub_Division_Name": sub_Division_Name ?? details.sub_Division_Name,
        "New_Home_Sales_Agent":
            new_Home_Sales_Agent ?? details.new_Home_Sales_Agent,
        "Sales_Agent_Office_Phone_Area_Code": agentPhoneOffice?.areaCode ??
            details.salesAgentOfficePhone?.areaCode,
        "Sales_Agent_Office_Phone_Prefix":
            agentPhoneOffice?.prefix ?? details.salesAgentOfficePhone?.prefix,
        "Sales_Agent_Office_Phone_Number":
            agentPhoneOffice?.number ?? details.salesAgentOfficePhone?.number,
        "Sales_Agent_Cell_Phone_Area_Code":
            agentPhoneCell?.areaCode ?? details.salesAgentCellPhone?.areaCode,
        "Sales_Agent_Cell_Phone_Prefix":
            agentPhoneCell?.prefix ?? details.salesAgentCellPhone?.prefix,
        "Sales_Agent_Cell_Phone_Number":
            agentPhoneCell?.number ?? details.salesAgentCellPhone?.number,
        "Sales_Agent_Email_Address":
            sales_Agent_Email_Address ?? details.salesAgentEmail,
        "Date_PreApproved": date_PreApproved ?? details.date_Pre_Approved,
        "Real_Deal_Date": real_Deal_Date ?? details.real_Deal_Date,
        "Lead_Date": lead_Date ?? details.lead_date,
        "Lead_View": lead_View ?? details.lead_View,
        "Last_Contact": details.lastContact,
        "Last_Attempt": details.lastAttempt,
      };
    } else {
      _data = {
        "Contact_ID": contact_ID,
        "Borrower_First_Name": borrower_First_Name,
        "Borrower_Last_Name": borrower_Last_Name,
        "Borrower_Birthdate": borrower_Birthdate ?? defaultDate,
        "Borrower_Phone1_Type": borrower_Phone1_Type ?? "",
        "Borrower_Phone1_Area_Code": borrowerPhone1?.areaCode ?? "",
        "Borrower_Phone1_Prefix": borrowerPhone1?.prefix ?? "",
        "Borrower_Phone1_Number": borrowerPhone1?.number ?? "",
        "Borrower_Phone2_Type": borrower_Phone2_Type ?? "",
        "Borrower_Phone2_Area_Code": borrowerPhone2?.areaCode ?? "",
        "Borrower_Phone2_Prefix": borrowerPhone2?.prefix ?? "",
        "Borrower_Phone2_Number": borrowerPhone2?.number ?? "",
        "Borrower_Email_Address": borrower_Email_Address ?? "",
        "Borrower_Contact_Preference": borrower_Contact_Preference ?? "",
        "Borrower_Credit_Report_Auth": borrower_Credit_Report_Auth ?? "",
        "Borrower_Credit_Score": borrower_Credit_Score ?? 0,
        "CoBorrower_First_Name": coBorrower_First_Name ?? "",
        "CoBorrower_Last_Name": coBorrower_Last_Name ?? "",
        "CoBorrower_Birthdate": coBorrower_Birthdate ?? defaultDate,
        "CoBorrower_Phone1_Type": coBorrower_Phone1_Type ?? "",
        "CoBorrower_Phone1_Area_Code": coBorrowerPhone1?.areaCode ?? "",
        "CoBorrower_Phone1_Prefix": coBorrowerPhone1?.prefix ?? "",
        "CoBorrower_Phone1_Number": coBorrowerPhone1?.number ?? "",
        "CoBorrower_Phone2_Type": coBorrower_Phone2_Type ?? "",
        "CoBorrower_Phone2_Area_Code": coBorrowerPhone2?.areaCode ?? "",
        "CoBorrower_Phone2_Prefix": coBorrowerPhone2?.prefix ?? "",
        "CoBorrower_Phone2_Number": coBorrowerPhone2?.number ?? "",
        "CoBorrower_Email_Address": coBorrower_Email_Address ?? "",
        "CoBorrower_Contact_Preference": coBorrower_Contact_Preference ?? "",
        "CoBorrower_Credit_Report_Auth": coBorrower_Credit_Report_Auth ?? "",
        "CoBorrower_Credit_Score": coBorrower_Credit_Score ?? 0,
        "Hot_Lead": hot_Lead ?? "",
        "Next_Step": next_Step ?? "",
        "Next_Action_Date": next_Action_Date ?? defaultDate,
        "Disposition": disposition ?? "",
        "Assign_Lead_To": assign_Lead_To ?? "",
        "Initial_Appointment_Date": initial_Appointment_Date ?? defaultDate,
        "Appointment_Type": appointment_Type ?? "",
        "Mailing_Address1": mailing_Address1 ?? "",
        "Mailing_Address2": mailing_Address2 ?? "",
        "Mailing_City": mailing_City ?? "",
        "Mailing_State": mailing_State ?? "",
        "Mailing_Zip": mailing_Zip ?? "",
        "Closing_Date": closing_Date ?? defaultDate,
        "Interest_Rate": interest_Rate ?? 0,
        "Loan_Amount": loan_Amount ?? 0,
        "Appraised_Value": appraised_Value ?? 0,
        "Purchase_Price": purchase_Price ?? 0,
        "Monthly_Payments": monthly_Payments ?? 0,
        "Target_Refi_Rate": target_Refi_Rate ?? 0,
        "Loan_Type": loan_Type ?? "",
        "Property_Type": property_City ?? "",
        "Occupancy": occupancy ?? "",
        "Loan_Number": loan_Number ?? "",
        "Loan_Purpose": loan_Purpose ?? "",
        "Property_City": property_City ?? "",
        "Property_State": property_State ?? "",
        "Time_Frame": time_Frame ?? "",
        "Property_Value": property_Value ?? 0,
        "Request_Loan_Amount": request_Loan_Amount ?? 0,
        "Request_Purchase_Price": request_Purchase_Price ?? 0,
        "Down_Payment": down_Payment ?? "",
        "Is_CoBorrower": is_CoBorrower ?? "",
        "Is_Military_Service": is_Military_Service ?? "",
        "Details_Complete": details_Complete ?? "",
        "Documentation_Complete": documentation_Complete ?? "",
        "Encompass_File_Created": encompass_File_Created ?? "",
        "Builder_Loan": builder_Loan ?? "",
        "Lead_Source": lead_Source ?? "",
        "Lead_Source_Type": lead_Source_Type ?? "",
        "Lead_Source_Name": lead_Source_Name ?? "",
        "Lead_Source_Office_Phone_Area_Code": leadPhoneOffice?.areaCode ?? "",
        "Lead_Source_Office_Phone_Prefix": leadPhoneOffice?.prefix ?? "",
        "Lead_Source_Office_Phone_Number": leadPhoneOffice?.number ?? "",
        "Lead_Source_Cell_Phone_Area_Code": leadPhoneCell?.areaCode ?? "",
        "Lead_Source_Cell_Phone_Prefix": leadPhoneCell?.prefix ?? "",
        "Lead_Source_Cell_Phone_Number": leadPhoneCell?.number ?? "",
        "Lead_Source_Email_Address": lead_Source_Email_Address ?? "",
        "Lead_Source_Contact_ID": lead_Source_Contact_ID ?? "",
        "Sub_Division_Name": sub_Division_Name ?? "",
        "New_Home_Sales_Agent": new_Home_Sales_Agent ?? "",
        "Sales_Agent_Office_Phone_Area_Code": agentPhoneOffice?.areaCode ?? "",
        "Sales_Agent_Office_Phone_Prefix": agentPhoneOffice?.prefix ?? "",
        "Sales_Agent_Office_Phone_Number": agentPhoneOffice?.number ?? "",
        "Sales_Agent_Cell_Phone_Area_Code": agentPhoneCell?.areaCode ?? "",
        "Sales_Agent_Cell_Phone_Prefix": agentPhoneCell?.prefix ?? "",
        "Sales_Agent_Cell_Phone_Number": agentPhoneCell?.number ?? "",
        "Sales_Agent_Email_Address": sales_Agent_Email_Address ?? "",
        "Date_PreApproved": date_PreApproved ?? defaultDate,
        "Real_Deal_Date": real_Deal_Date ?? defaultDate,
        "Lead_Date":
            lead_Date ?? DateFormat.yMd().add_jms().format(DateTime.now()),
        "Lead_View": lead_View ?? "Active",
        "Total_Count": total_count ?? ""
      };
    }
    Map<String, dynamic> result;

    if (update) {
      if (updateID != null && updateID.isNotEmpty) {
        result = await getJsonMap("Core/Leads/$updateID",
            body: _data, extract: null, method: 'PUT');
      } else {
        throw Exception("updateID must be provided while updating a CoreLead");
      }
    } else if (attach) {
      if (attachContact != null && attachContact.isNotEmpty) {
        result = await getJsonMap("Core/leads/attach/$attachContact",
            body: _data, extract: null);
      } else {
        throw Exception(
            'attachContact must be provided while attaching a ContactID');
      }
    } else {
      //await addContact(firstName: borrower_First_Name, lastName: borrower_Last_Name);
      result = await getJsonMap("Core/leads/add", body: _data, extract: null);
    }
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception(
            "Failed to add lead '${borrower_First_Name}': ${reason}");
      } else {
        throw Exception("Failed to add lead '${borrower_First_Name}'");
      }
    }
  }

  @override
  Future<List<CoreLeadNote>> listCoreLeadNotes({String id}) async {
    var notes = <CoreLeadNote>[];
    List<dynamic> _data;

    _data = await getJsonList("Core/details/$id/notes");
    //if(_data == null) throw Exception("Lead Not Found");

    for (var note in _data) {
      notes.add(CoreLeadNote(this, note));
    }
    print('test');
    return notes;
  }

  @override
  Future addCoreLeadFollowUp(String id,
      {String title,
      DateTime date,
      int type = 0,
      String description = "",
      bool dismissOnChange = true,
      String alertType,
      int alertOffset = 0,
      String userNumber,
      String prospectNumber}) async {
    var _data = {
      "task_type": type ?? 0,
      "title": title,
      "description": description ?? "",
      "date_time": (date ?? DateTime.now()).toString(),
      "dismiss_on_change": dismissOnChange ?? true,
      "alert_type": alertType ?? "",
      "alert_offset": alertOffset ?? 0,
      "prospect_number": userNumber,
      "user_number": prospectNumber
    };
    var result = await getJsonMap("Core/leads/$id/follow_up",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add follow up '${title}': ${reason}");
      } else {
        throw Exception("Failed to add follow up '${title}'");
      }
    }
  }

  @override
  Future<DashboardStats> getStats() async {
    return DashboardStats(this, await getJsonMap("dashboard/user_statistics"));
  }

  @override
  Future<LoanProgress> getLoanStatus(String id) async {
    return LoanProgress(this, await getJsonMap("loans/info/$id/status"));
  }

  @override
  Future<List<LoanContact>> listLoanContacts(String id) async {
    var items = <LoanContact>[];

    List<dynamic> result;

    result = await getJsonList("loans/info/$id/contacts");

    for (var item in result) {
      items.add(LoanContact(this, item));
    }

    return items;
  }

  @override
  Future<List<LoanHistory>> listLoanHistory(String id) async {
    var items = <LoanHistory>[];

    List<dynamic> result;

    result = await getJsonList("contacts/details/$id/loan_history");

    for (var item in result) {
      items.add(LoanHistory(this, item));
    }

    return items;
  }

  @override
  Future<List<String>> listContactLogResponses() async {
    var items = <String>[];

    List<dynamic> result;

    result = await getJsonList("contacts/details/log_responses");

    for (var item in result) {
      items.add(item as String);
    }

    return items;
  }

  @override
  Future<List<String>> listLeadLogResponses() async {
    var items = <String>[];

    List<dynamic> result;

    result = await getJsonList("leads/details/log_responses");

    for (var item in result) {
      items.add(item as String);
    }

    return items;
  }

  @override
  Future<List<CartItem>> listCartItems() async {
    var items = <CartItem>[];

    List<dynamic> result;

    result = await getJsonList("cart/items");

    for (var item in result) {
      items.add(CartItem(this, item));
    }

    return items;
  }

  @override
  Future<List<UnifyAlert>> listUnifyAlerts(
      {String type, String priority, DateTime date}) async {
    var items = <UnifyAlert>[];

    List<dynamic> result;

    String _search = "";
    if (type != null && type.isNotEmpty && type != "All") {
      if (_search.isNotEmpty) {
        _search += "&type=$type";
      } else {
        _search += "?type=$type";
      }
    }
    if (priority != null && priority.isNotEmpty && priority != "All") {
      if (_search.isNotEmpty) {
        _search += "&priority=$priority";
      } else {
        _search += "?priority=$priority";
      }
    }

    if (date != null) {
      String _date =
          DateFormat(_calendarDateFormat).format(date ?? DateTime.now());
      result = await getJsonList("account/alerts/$_date$_search");
    } else {
      result = await getJsonList("account/alerts$_search");
    }

    for (var item in result) {
      items.add(UnifyAlert(this, item));
    }

    return items;
  }

  @override
  Future deleteUnifyAlert(String id) async {
    var result =
        await getJsonMap("account/alerts/$id", extract: null, method: "DELETE");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete alert '${id}': ${reason}");
      } else {
        throw Exception("Failed to delete alert '${id}'");
      }
    }
  }

  @override
  Future<CartOrder> getCartOrder() async {
    var _data = await getJsonMap("cart/order");
    if (_data == null) throw Exception("Order Not Found");
    List<CartItem> _items = await listCartItems();
    _data["Cart_Items"] = _items?.map((i) => i?.json)?.toList() ?? [];
    return CartOrder(this, _data);
  }

  @override
  Future deleteCartItem(String id) async {
    var result =
        await getJsonMap("cart/items/$id", extract: null, method: "DELETE");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete cart item '${id}': ${reason}");
      } else {
        throw Exception("Failed to delete cart item '${id}'");
      }
    }
  }

  @override
  Future deleteCartCard(String id) async {
    var result =
        await getJsonMap("cart/cards/$id", extract: null, method: "DELETE");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete cart card '${id}': ${reason}");
      } else {
        throw Exception("Failed to delete cart card '${id}'");
      }
    }
  }

  @override
  Future deleteContact(String id) async {
    var result = await getJsonMap("contacts/details/$id",
        extract: null, method: "DELETE");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete contact '${id}': ${reason}");
      } else {
        throw Exception("Failed to delete contact '${id}'");
      }
    }
  }

  @override
  Future deleteLead(String id) async {
    var result =
        await getJsonMap("leads/details/$id", extract: null, method: "DELETE");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete lead '${id}': ${reason}");
      } else {
        throw Exception("Failed to delete lead '${id}'");
      }
    }
  }

  @override
  Future deleteTask(String id) async {
    var result =
        await getJsonMap("calendar/$id", extract: null, method: "DELETE");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete task '${id}': ${reason}");
      } else {
        throw Exception("Failed to delete task '${id}'");
      }
    }
  }

  @override
  Future batchDeleteContacts(List<Contact> contacts) async {
    var _data = contacts?.map((i) => i?.json)?.toList() ?? [];
    var result = await sendJsonList("contacts/batch/delete",
        body: _data, extract: null, method: "POST");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete contacts '${contacts}': ${reason}");
      } else {
        throw Exception("Failed to delete contacts '${contacts}'");
      }
    }
  }

  @override
  Future batchDeleteLeads(List<Lead> leads) async {
    var _data = leads?.map((i) => i?.json)?.toList() ?? [];
    var result = await sendJsonList("leads/batch/delete",
        body: _data, extract: null, method: "POST");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete leads '${leads}': ${reason}");
      } else {
        throw Exception("Failed to delete leads '${leads}'");
      }
    }
  }

  @override
  Future batchRestoreContacts(List<Contact> contacts) async {
    var _data = contacts?.map((i) => i?.json)?.toList() ?? [];
    var result = await sendJsonList("contacts/batch/restore",
        body: _data, extract: null, method: "POST");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to restore contacts '${contacts}': ${reason}");
      } else {
        throw Exception("Failed to restore contacts '${contacts}'");
      }
    }
  }

  @override
  Future batchRestoreLeads(List<Lead> leads) async {
    var _data = leads?.map((i) => i?.json)?.toList() ?? [];
    var result = await sendJsonList("leads/batch/restore",
        body: _data, extract: null, method: "POST");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to restore leads '${leads}': ${reason}");
      } else {
        throw Exception("Failed to restore leads '${leads}'");
      }
    }
  }

  @override
  Future deleteDashboardAlert(String id) async {
    var result =
        await getJsonMap("dashboard/$id", extract: null, method: "DELETE");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete alert '${id}': ${reason}");
      } else {
        throw Exception("Failed to delete alert '${id}'");
      }
    }
  }

  @override
  Future restoreDashboardAlert(String id) async {
    var result =
        await getJsonMap("dashboard/$id", extract: null, method: "PUT");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to restore alert '${id}': ${reason}");
      } else {
        throw Exception("Failed to restore alert '${id}'");
      }
    }
  }

  @override
  Future batchDeleteDashboardAlerts(List<DashboardAlert> alerts) async {
    var _data = alerts?.map((i) => i?.json)?.toList() ?? [];
    var result = await sendJsonList("dashboard/batch/restore",
        body: _data, extract: null, method: "POST");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete alerts '${alerts}': ${reason}");
      } else {
        throw Exception("Failed to delete alerts '${alerts}'");
      }
    }
  }

  @override
  Future batchRestoreDashboardAlerts(List<DashboardAlert> alerts) async {
    var _data = alerts?.map((i) => i?.json)?.toList() ?? [];
    var result = await sendJsonList("dashboard/batch/restore",
        body: _data, extract: null, method: "POST");
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to restore alerts '${alerts}': ${reason}");
      } else {
        throw Exception("Failed to restore alerts '${alerts}'");
      }
    }
  }

  @override
  Future addContactGroup(String name) async {
    var _data = {"name": name ?? ""};
    var result = await getJsonMap("contacts/contact_groups/add",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add group '${name}': ${reason}");
      } else {
        throw Exception("Failed to add group '${name}'");
      }
    }
  }

  @override
  Future editContactGroup(String id, String name) async {
    var _data = {"name": name ?? ""};
    var result = await getJsonMap("contacts/contact_groups/$id",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to edit group '${name}': ${reason}");
      } else {
        throw Exception("Failed to edit group '${name}'");
      }
    }
  }

  @override
  Future addLeadGroup(String name) async {
    var _data = {"name": name ?? ""};
    var result =
        await getJsonMap("leads/lead_groups/add", body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add group '${name}': ${reason}");
      } else {
        throw Exception("Failed to add group '${name}'");
      }
    }
  }

  @override
  Future editLeadGroup(String id, String name) async {
    var _data = {"name": name ?? ""};
    var result =
        await getJsonMap("leads/lead_groups/$id", body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to edit group '${name}': ${reason}");
      } else {
        throw Exception("Failed to edit group '${name}'");
      }
    }
  }

  @override
  Future deleteContactGroup(String id) async {
    var result = await getJsonMap("contacts/contact_groups/$id",
        method: "DELETE", extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete group '${id}': ${reason}");
      } else {
        throw Exception("Failed to delete group '${id}'");
      }
    }
  }

  @override
  Future deleteLeadGroup(String id) async {
    var result = await getJsonMap("leads/lead_groups/$id",
        method: "DELETE", extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to delete group '${id}': ${reason}");
      } else {
        throw Exception("Failed to delete group '${id}'");
      }
    }
  }

  @override
  Future deleteMarketingCampiagn(String id) async {
    throw Exception("This platform is not supported.");
  }

  @override
  Future<UnifyCache> getCache({bool all = false}) async {
    AccountInfo _info = await getInfo();
    List<Contact> _contacts = await listContacts();
    List<Lead> _leads = await listLeads();
    List<ContactGroup> _contactGroups = await listContactGroups();
    List<LeadGroup> _leadGroups = await listLeadGroups();
    List<Loan> _loansAll = await listLoans();
    List<DashboardAlert> _dashboardOpportunity = await listOpportunitiyAlerts();
    List<DashboardAlert> _dashboardCustomer = await listCustomerAlerts();
    List<DashboardAlert> _dashboardHotProspects = await listHotProspectAlerts();
    List<DashboardAlert> _dashboardPipeline = await listPipelineAlerts();
    List<Task> _tasksAll = await listTasksAll();
    List<Video> _videosAll = await listVideosAll();
    List<MarketingSection> _marketingTree = await listMarketingSection();
    List<MarketingCampaign> _marketingCampaigns =
        await listMarketingCampaigns();

    Map<String, dynamic> _data = {
      "account_info": _info?.json,
      "contacts": _contacts?.map((i) => i?.json)?.toList() ?? [],
      "leads": _leads?.map((i) => i?.json)?.toList() ?? [],
      "contact_groups": _contactGroups?.map((i) => i?.json)?.toList() ?? [],
      "lead_groups": _leadGroups?.map((i) => i?.json)?.toList() ?? [],
      "dashboard_opportunity":
          _dashboardOpportunity?.map((i) => i?.json)?.toList() ?? [],
      "dashboard_customer":
          _dashboardCustomer?.map((i) => i?.json)?.toList() ?? [],
      "dashboard_hotprospects":
          _dashboardHotProspects?.map((i) => i?.json)?.toList() ?? [],
      "dashboard_pipeline":
          _dashboardPipeline?.map((i) => i?.json)?.toList() ?? [],
      "videos_all": _videosAll?.map((i) => i?.json)?.toList() ?? [],
      "loans_all": _loansAll?.map((i) => i?.json)?.toList() ?? [],
      "marketing_tree": _marketingTree?.map((i) => i?.json)?.toList() ?? [],
      "marketing_campaigns":
          _marketingCampaigns?.map((i) => i?.json)?.toList() ?? [],
      "last_updated": DateTime.now(),
    };

    if (all) {
      final _tasksContacts = await listTasksContacts();
      final _tasksLeads = await listTasksLeads();
      final _loansMyPipeline = await listLoans(folder: LoansFolder.mypipeline);
      final _loansProspects = await listLoans(folder: LoansFolder.prospects);
      final _loansClosed = await listLoans(folder: LoansFolder.closed);
      final _loansOther = await listLoans(folder: LoansFolder.other);
      final _videosMy = await listVideosMyVideos();
      final _videosShared = await listVideosShared();
      final _videosArchived = await listVideosArchived();
      final _unifyAlerts = await listUnifyAlerts();

      _data["tasks_all"] = _tasksAll?.map((i) => i?.json)?.toList() ?? [];
      _data["tasks_contacts"] =
          _tasksContacts?.map((i) => i?.json)?.toList() ?? [];
      _data["tasks_leads"] = _tasksLeads?.map((i) => i?.json)?.toList() ?? [];
      _data["loans_mypipeline"] =
          _loansMyPipeline?.map((i) => i?.json)?.toList() ?? [];
      _data["loans_prospects"] =
          _loansProspects?.map((i) => i?.json)?.toList() ?? [];
      _data["loans_closed"] = _loansClosed?.map((i) => i?.json)?.toList() ?? [];
      _data["loans_other"] = _loansOther?.map((i) => i?.json)?.toList() ?? [];
      _data["videos_all"] = _videosMy?.map((i) => i?.json)?.toList() ?? [];
      _data["videos_shared"] =
          _videosShared?.map((i) => i?.json)?.toList() ?? [];
      _data["videos_archived"] =
          _videosArchived?.map((i) => i?.json)?.toList() ?? [];
      _data["unify_alerts"] = _unifyAlerts?.map((i) => i?.json)?.toList() ?? [];
    }
    return UnifyCache(this, _data);
  }

  final String _kalturaUrl = "https://www.kaltura.com/api_v3/service/";

  @override
  Future<List<Video>> listVideosAll(
      {int page = 1, int rows = 100, String query = ""}) async {
    var _info = await getInfo();
    if (!_info.videoEnabled) throw Exception("Unify Video Not Set Up");
    String _session = await getKalturaSession(_info?.kalturaId);
    int count = (page * rows).round();

    var items = <Video>[];

    List<dynamic> result;

    String _base =
        "$_kalturaUrl\media/action/list?ks=$_session&format=1&pager%5BpageSize%5D=$count";
    String _section = "&filter[orderBy]=CREATED_AT_ASC";

    if (query != null && query.isNotEmpty) {
      final _query = Uri.encodeQueryComponent(query);
      result = await getJsonList(
        null,
        baseUrl: endpoints.kalturaUrl
            .resolve("$_base$_section&filter[nameLike]=$_query"),
        extract: "objects",
      );
    } else {
      result = await getJsonList(
        null,
        baseUrl: endpoints.kalturaUrl.resolve("$_base$_section"),
        extract: "objects",
      );
    }

    for (var item in result) {
      items.add(Video(this, item));
    }

    return items;
  }

  @override
  Future<List<Video>> listVideosMyVideos(
      {int page = 1,
      int rows = 100,
      String query = "",
      bool showUnapproved = false}) async {
    String _section = '';
    var _info = await getInfo();
    if (!_info.videoEnabled) throw Exception("Unify Video Not Set Up");
    String _session = await getKalturaSession(_info?.kalturaId);
    int count = (page * rows).round();

    var items = <Video>[];

    List<dynamic> result;
    String _base =
        "media/action/list?ks=$_session&format=1&pager%5BpageSize%5D=$count";

    if (showUnapproved) {
      _section =
          "&filter%5BcategoriesNotContains%5D=Archived,UnapprovedShared,ApprovedShared,Shared&filter%5BorderBy%5D=CREATED_AT_ASC";
    } else {
      _section =
          "&filter%5BcategoriesNotContains%5D=Archived,Unapproved,UnapprovedShared,ApprovedShared,Shared&filter%5BorderBy%5D=CREATED_AT_ASC";
    }

    if (query != null && query.isNotEmpty) {
      final _query = Uri.encodeQueryComponent(query);
      result = await getJsonList(
        null,
        baseUrl: endpoints.kalturaUrl
            .resolve("$_base$_section&filter[nameLike]=$_query"),
        extract: "objects",
      );
    } else {
      result = await getJsonList(
        null,
        baseUrl: endpoints.kalturaUrl.resolve("$_base$_section"),
        extract: "objects",
      );
    }

    for (var item in result) {
      items.add(Video(this, item));
    }

    return items;
  }

  @override
  Future<List<Video>> listVideosArchived(
      {int page = 1, int rows = 100, String query = ""}) async {
    var _info = await getInfo();
    if (!_info.videoEnabled) throw Exception("Unify Video Not Set Up");
    String _session = await getKalturaSession(_info?.kalturaId);
    int count = (page * rows).round();

    var items = <Video>[];

    List<dynamic> result;
    String _base =
        "media/action/list?ks=$_session&format=1&pager%5BpageSize%5D=$count";
    String _section =
        "&filter%5BorderBy%5D=CREATED_AT_DESC&filter%5BcategoriesMatchOr%5D=Archived";

    if (query != null && query.isNotEmpty) {
      final _query = Uri.encodeQueryComponent(query);
      result = await getJsonList(
        null,
        baseUrl: endpoints.kalturaUrl
            .resolve("$_base$_section&filter[nameLike]=$_query"),
        extract: "objects",
      );
    } else {
      result = await getJsonList(
        null,
        baseUrl: endpoints.kalturaUrl.resolve("$_base$_section"),
        extract: "objects",
      );
    }

    for (var item in result) {
      items.add(Video(this, item));
    }

    return items;
  }

  @override
  Future<List<Video>> listVideosShared(
      {int page = 1, int rows = 100, String query = ""}) async {
    var _info = await getInfo();
    if (!_info.videoEnabled) throw Exception("Unify Video Not Set Up");
    String _session = await getKalturaSession(null, admin: true);
    String _companyId = _info?.companyId ?? "";
    int count = (page * rows).round();

    var items = <Video>[];

    List<dynamic> result;
    String _base =
        "media/action/list?ks=$_session&format=1&pager%5BpageSize%5D=$count";
    String _section =
        "&filter%5BorderBy%5D=CREATED_AT_DESC&filter%5BtagsLike%5D=$_companyId&filter[categoriesMatchOr]=UnapprovedShared,ApprovedShared,Shared";

    if (query != null && query.isNotEmpty) {
      final _query = Uri.encodeQueryComponent(query);
      result = await getJsonList(
        null,
        baseUrl: endpoints.kalturaUrl
            .resolve("$_base$_section&filter[nameLike]=$_query"),
        extract: "objects",
      );
    } else {
      result = await getJsonList(
        null,
        baseUrl: endpoints.kalturaUrl.resolve("$_base$_section"),
        extract: "objects",
      );
    }

    for (var item in result) {
      items.add(Video(this, item));
    }

    return items;
  }

  @override
  Future editVideoInfo(String id, {String name, String description}) async {
    var _info = await getInfo();
    var _video = await getVideoDetails(id);
    if (!_info.videoEnabled) throw Exception("Unify Video Not Set Up");
    String _session = await getKalturaSession(_info?.kalturaId);
    String _companyId = _info?.companyId ?? "";

    var _formName = Uri.encodeQueryComponent(r'mediaEntry[name]');
    var _formDesription = Uri.encodeQueryComponent(r'mediaEntry[description]');
    var _formTags = Uri.encodeQueryComponent(r'mediaEntry[tags]');
    var _name = Uri.encodeQueryComponent(
        (name ?? _video?.name ?? "")?.trim()?.toString());
    var _desc = Uri.encodeQueryComponent(
        (description ?? _video?.description ?? "")?.trim()?.toString());

    return await getJsonMap(
      null,
      baseUrl: endpoints.kalturaUrl.resolve(
          "media/action/update?ks=$_session&format=1&entryId=$id&$_formName=$_name&$_formDesription=$_desc&$_formTags=$_companyId"),
    );
  }

  @override
  Future<int> getVideoTotalViews() async {
    var _info = await getInfo();
    if (!_info.videoEnabled) throw Exception("Unify Video Not Set Up");
    List<Video> _videos = await listVideosAll(rows: 1000000);
    num _count = 0;
    for (Video _video in _videos) {
      _count += _video?.views ?? 0;
    }
    return _count?.toInt() ?? 0;
  }

  @override
  Future<String> getVideoPreview(
    String id, {
    String name,
    String body,
    String template,
  }) async {
    List<String> _templates = await listTemplates();
    if (!_templates.contains((template ?? "MESSAGE").toUpperCase()))
      throw Exception("Not a Valid Template");
    VideoDetails _info = await getVideoDetails(id);
    AccountInfo _user = await getInfo();
    return endpoints.userApiUrl
        .resolve(
            "email/preview?userID=${_user?.id}&videoID=$id&body=${body ?? _info?.description ?? ""}&name=${name ?? _user?.fullName}&template=${(template ?? "MESSAGE").toUpperCase()}")
        .toString()
        .replaceAll("api.", "www.")
        .replaceAll("mobile.", "www.");
  }

  @override
  Future startVideoQuickSend(
    String id, {
    bool cc,
    @required List<String> emails,
    @required String subject,
    @required String body,
    String name,
    String template,
  }) async {
    List<String> _templates = await listTemplates();
    if (!_templates.contains((template ?? "MESSAGE").toUpperCase()))
      throw Exception("Not a Valid Template");
    VideoDetails _info = await getVideoDetails(id);
    AccountInfo _user = await getInfo();
    var _data = {
      "cc": cc ?? false,
      "emails": emails ?? [],
      "name": name ?? _user?.fullName,
      "template": (template ?? "MESSAGE").toUpperCase(),
      "video_id": id,
      "subject": name ?? _info?.name ?? "",
      "body": body ?? _info?.description ?? "",
    };
    var result = await getJsonMap("email/send", body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to start quick send '${id}': ${reason}");
      } else {
        throw Exception("Failed to start quick send '${id}'");
      }
    }
  }

  @override
  Future<String> getVideoShareUrl(String id,
      {String template = "MESSAGE"}) async {
    AccountInfo _info = await getInfo();
    String company = _info?.companyName ?? "Unify";
    return endpoints.userApiUrl
        .resolve(
            "VideoPlayerTemplate.aspx?entryid=$id&company=$company&temp=$template")
        .toString()
        .replaceAll("api.", "www.")
        .replaceAll("mobile.", "www.");
  }

  @override
  Future<VideoDetails> getVideoDetails(String id) async {
    AccountInfo _info = await getInfo();
    if (_info?.kalturaId == null) throw Exception("Unify Video Not Set Up");
    String _session = await getKalturaSession(_info?.kalturaId);
    Map<String, dynamic> _data = await getJsonMap(
      null,
      baseUrl: endpoints.kalturaUrl.resolve(
          "media/action/get?ks=$_session&format=1&entryId=$id&version=-1"),
      extract: null,
    );
    if (_data == null) throw Exception("Video Not Found");
    return VideoDetails(this, _data);
  }

  Future<String> getKalturaSession(String kalturaID,
      {bool admin = false}) async {
    if (admin) {
      String _username = "sl@crossmedia-llc.com";
      String _password = "Video\$2017";
      var result = await postHttp(
        null,
        baseUrl: endpoints.kalturaUrl.resolve(
            "user/action/loginbyloginid?loginId=$_username&password=$_password&format=1"),
        extract: null,
        raw: true,
      );
      if (result == null) throw Exception("Error Getting Session");
      return result.toString().replaceAll("\"", "");
    }

    String _secret = "027deb191007aa0565a90e21d9c07cf6";
    String _partnerId = "2336561";
    String _expiry = "86400";
    var result = await sendHttpRequest(
      null,
      baseUrl: endpoints.kalturaUrl.resolve(
          "session/action/start?secret=$_secret&userId=$kalturaID&type=0&partnerId=$_partnerId&expiry=$_expiry&format=1"),
      extract: null,
    );

    if (result == null) throw Exception("Error Getting Session");
    return result.toString().replaceAll("\"", "");
  }

  @override
  Future<List<String>> listTemplates() async {
    var items = <String>[];

    List<dynamic> result;

    result = await getJsonList("email/templates");

    for (var item in result) {
      items.add(item as String);
    }

    return items;
  }

  @override
  Future<List<MarketingSection>> listMarketingSection() async {
    var items = <MarketingSection>[];

    List<dynamic> result;

    result = await getJsonList("marketing/types/tree");

    for (var item in result) {
      items.add(MarketingSection(this, item));
    }

    items?.sort((MarketingSection a, MarketingSection b) {
      return a.name.toLowerCase().compareTo(b.name.toLowerCase());
    });

    for (MarketingSection type in items) {
      if (type?.items != null)
        for (MarketingCategory category in type?.items) {
          Map<String, dynamic> _categoryData = category?.json;
          _categoryData["marketing_type"] = type.name;
          MarketingCategory _category = MarketingCategory(this, _categoryData);
          if (category?.items != null)
            for (MarketingSubCategory subcategory in category?.items) {
              Map<String, dynamic> _subCategoryData = subcategory?.json;
              _subCategoryData["marketing_type"] = type.name;
              _subCategoryData["marketing_category"] = category.name;
              MarketingSubCategory _subCategory =
                  MarketingSubCategory(this, _categoryData);
              subcategory = _subCategory;
            }
          category = _category;
        }
    }

    return items;
  }

  @override
  Future<List<MarketingMaterial>> listMarketingMaterial(
      {String type,
      String category,
      String subCategory,
      String query = ""}) async {
    var items = <MarketingMaterial>[];

    List<dynamic> result;

    if (query != null && query.isNotEmpty) {
      String _url = "";
      if ((query?.contains("cobrand") ?? false) ||
          (query?.toLowerCase()?.contains("co brand") ?? false)) {
        _url =
            "marketing/material/search?name=$query&cobrand=True&Search_All=true";
      } else {
        if (subCategory != null && subCategory.isNotEmpty) {
          _url =
              "marketing/material/search?type=$type&category=$category&sub_category=$subCategory&name=$query&Search_All=false";
        } else if (category != null && category.isNotEmpty) {
          _url =
              "marketing/material/search?type=$type&category=$category&name=$query&Search_All=false";
        } else if (type != null && type.isNotEmpty) {
          _url =
              "marketing/material/search?type=$type&name=$query&Search_All=false";
        } else {
          _url = "marketing/material/search?name=$query&Search_All=false";
        }
      }
      result = await getJsonList(_url);
    } else {
      String _url = "";
      if (subCategory != null && subCategory.isNotEmpty) {
        _url = "marketing/material/$type/$category/$subCategory";
      } else if (category != null && category.isNotEmpty) {
        _url = "marketing/material/$type/$category";
      } else if (type != null && type.isNotEmpty) {
        _url = "marketing/material/$type";
      } else {
        _url = "marketing/material";
      }
      result = await getJsonList(_url);
    }

    for (var item in result) {
      items.add(MarketingMaterial(this, item));
    }

    return items;
  }

  @override
  Future<List<MarketingCampaign>> listMarketingCampaigns(
      {int page = 1, int rows = 100}) async {
    var items = <MarketingCampaign>[];

    List<dynamic> result;

    result = await getJsonList("marketing/$rows/$page");

    for (var item in result) {
      items.add(MarketingCampaign(this, item));
    }

    return items;
  }

  @override
  Future<MarketingMaterialDetails> getMarketingDetails(String id) async {
    var _data = await getJsonMap("marketing/details/$id");
    if (_data == null) throw Exception("Marketing Material Not Found");
    return MarketingMaterialDetails(this, _data);
  }

  @override
  Future<MarketingPreview> getMarketingPreview(String id) async {
    var _data = await getJsonMap("marketing/preview/$id");
    if (_data == null) throw Exception("Marketing Material Preview Not Found");
    return MarketingPreview(this, _data);
  }

  @override
  Future<List<ImpUsers>> getImpersinationList() async {
    var items = <ImpUsers>[];

    List<dynamic> result;

    result = await getJsonList("Account/impersonate/list");

    for (var item in result) {
      items.add(ImpUsers(this, item));
    }

    return items;
  }

  @override
  Future<List<ImpUsers>> searchImpersonationList(
      {int rows = 100,
      int page = 1,
      String lastName = '',
      String firstName = '',
      String username = ''}) async {
    final items = <ImpUsers>[];
    List<dynamic> result;

    if (username.isNotEmpty) {
      if (lastName.isNotEmpty) {
        if (firstName.isNotEmpty) {
          result = await getJsonList(
              'Account/impersonateQuery/list?rows=$rows&page=$page&username=$username&fast_Name=$firstName&last_Name=$lastName');
        }
        result = await getJsonList(
            'Account/impersonateQuery/list?rows=$rows&page=$page&username=$username&last_Name=$lastName');
      }
      result = await getJsonList(
          'Account/impersonateQuery/list?rows=$rows&page=$page&username=$username');
    } else {
      result = await getJsonList(
          'Account/impersonateQuery/list?rows=$rows&page=$page');
    }

    for (var item in result) {
      items.add(ImpUsers(this, item));
    }

    return items;
  }

  @override
  Future<UnifyJsonAccessToken> impersinateUser(String id) async {
    var _token = await getJsonMap("Account/impersonate/$id");
    if (_token == null) {
      throw Exception("Token not found");
    }
    return UnifyJsonAccessToken(_token);
  }

  @override
  Future<List<MarketingPartner>> listMarketingPartners() async {
    var items = <MarketingPartner>[];

    List<dynamic> result;

    result = await getJsonList("marketing/partners");

    for (var item in result) {
      items.add(MarketingPartner(this, item));
    }

    return items;
  }

  @override
  Future<MarketingValidation> getMarketingValidation(String id,
      {String partnerId,
      bool sendNow,
      DateTime startDate,
      List<String> contactIds,
      List<String> leadIds}) async {
    var _data = {
      "marketing_partner_id": partnerId ?? "",
      "send_now": sendNow ?? false,
      "start_date": (startDate ?? DateTime.now()).toString(),
      "contact_ids": contactIds ?? [],
      "lead_ids": leadIds ?? [],
    };
    return MarketingValidation(
        this, await getJsonMap("marketing/details/$id/validate", body: _data));
  }

  @override
  Future startMarketingCampaign(String id,
      {String partnerId,
      bool sendNow,
      DateTime startDate,
      List<String> contactIds,
      List<String> leadIds}) async {
    if ((contactIds == null || contactIds.isEmpty) &&
        (leadIds == null || leadIds.isEmpty)) {
      throw Exception("1 Contact / Lead Required");
    }
    var _data = {
      "marketing_partner_id": partnerId ?? "",
      "send_now": sendNow ?? false,
      "start_date": (startDate ?? DateTime.now()).toString(),
      "contact_ids": contactIds ?? [],
      "lead_ids": leadIds ?? [],
    };
    var result = await getJsonMap("marketing/details/$id/start_campaign",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to start campaign '${id}': ${reason}");
      } else {
        throw Exception("Failed to start campaign '${id}'");
      }
    }
  }

  @override
  Future addMarketingCampaignToCart(String id,
      {String partnerId,
      bool sendNow,
      DateTime startDate,
      List<String> contactIds,
      List<String> leadIds}) async {
    if ((contactIds == null || contactIds.isEmpty) &&
        (leadIds == null || leadIds.isEmpty)) {
      throw Exception("1 Contact / Lead Required");
    }
    var _data = {
      "marketing_partner_id": partnerId ?? "",
      "send_now": sendNow ?? false,
      "start_date": (startDate ?? DateTime.now()).toString(),
      "contact_ids": contactIds ?? [],
      "lead_ids": leadIds ?? [],
    };
    var result = await getJsonMap("marketing/details/$id/add_to_cart",
        body: _data, extract: null);
    if (result["Status"] != "Success") {
      var reason = result["Message"];
      if (reason is String && reason.trim().isNotEmpty) {
        throw Exception("Failed to add campaign to cart '${id}': ${reason}");
      } else {
        throw Exception("Failed to add campaign to cart '${id}'");
      }
    }
  }

  @override
  Future<AccountInfo> getUser(String id) async {
    var _data = await getJsonMap("account/$id");
    if (_data == null) throw Exception("User Not Found");
    return AccountInfo(this, _data);
  }

  @override
  Future<AccountInfo> getInfo() async {
    return AccountInfo(this, await getJsonMap("account/info"));
  }

  Future<num> getAccountBalance() async {
    var result =
        await sendJsonList("account/balance", extract: null, method: "GET");
    return result["Result"];
  }

  Future<Map<String, dynamic>> getJsonMap(
    String url, {
    Map<String, dynamic> body,
    String extract: "Result",
    bool standard: true,
    String method,
    Uri baseUrl,
  }) async {
    return (await sendHttpRequest(
      _apiUrl(url, standard),
      method: method,
      body: body,
      extract: extract,
      baseUrl: baseUrl,
    )) as Map<String, dynamic>;
  }

  Future<Map<String, dynamic>> sendJsonList(
    String url, {
    List<Map<String, dynamic>> body,
    List<String> strBody,
    String extract: "Result",
    bool standard: true,
    String method,
    Uri baseUrl,
  }) async {
    return (await sendHttpRequest(
      _apiUrl(url, standard),
      method: method,
      body: body ?? strBody,
      extract: extract,
      baseUrl: baseUrl,
    )) as Map<String, dynamic>;
  }

  Future<List<dynamic>> getJsonList(
    String url, {
    Map<String, dynamic> body,
    List<String> strBody,
    String extract: "Result",
    bool standard: true,
    String method,
    Uri baseUrl,
  }) async {
    return (await sendHttpRequest(
      _apiUrl(url, standard),
      method: method,
      body: body ?? strBody,
      extract: extract,
      baseUrl: baseUrl,
    )) as List<dynamic>;
  }

  Future<dynamic> sendHttpRequest(
    String url, {
    bool needsToken: true,
    String extract,
    dynamic body,
    String method,
    Uri baseUrl,
  });

  Future<dynamic> postHttp(
    String url, {
    String extract,
    Map<String, dynamic> body,
    Uri baseUrl,
    bool raw = false,
  });

  String _apiUrl(String path, bool standard) {
    // return standard ? "/api/1/$path" : path;
    return standard ? "/api/$path" : path;
  }
}
